{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Polynomial where {

import qualified Show;
import qualified Nat;
import qualified Product_Type;
import qualified ErrorMonad;
import qualified Term;
import qualified List;
import qualified HOL;
import qualified Util;
import qualified RedPair;

data Poly a = Poly Nat.Nat [(Nat.Nat, a)] deriving (Read, Show);

newtype RedPair a = Polo [(a, (Nat.Nat, [Nat.Nat]))] deriving (Read, Show);

genIndices :: forall a. Nat.Nat -> Integer -> [(Term.Term a [Char])];
genIndices (Nat.Suc n) i =
  Term.Var (List.append "x_" (Show.sshow_int i)) :
    Polynomial.genIndices n (i + 1);
genIndices Nat.Zero_nat uu = [];

genTerm :: forall a. a -> Nat.Nat -> Term.Term a [Char];
genTerm f n = Term.Fun f (Polynomial.genIndices n 0);

polyOne :: forall v. Polynomial.Poly v;
polyOne = Polynomial.Poly Nat.one_nat [];

lookupVar ::
  forall a. (Eq a) => a -> [(Nat.Nat, a)] -> Maybe (Nat.Nat, [(Nat.Nat, a)]);
lookupVar x ((c, y) : vp) =
  (if HOL.eqop x y then Just (c, vp)
    else (case Polynomial.lookupVar x vp of {
           Nothing -> Nothing;
           Just a -> let {
                       (d, vpp) = a;
                     } in Just (d, (c, y) : vpp);
         }));
lookupVar x [] = Nothing;

sumPolyVars ::
  forall a. (Eq a) => [(Nat.Nat, a)] -> [(Nat.Nat, a)] -> [(Nat.Nat, a)];
sumPolyVars ((c, x) : vp1) vp2 =
  (case Polynomial.lookupVar x vp2 of {
    Nothing -> (c, x) : Polynomial.sumPolyVars vp1 vp2;
    Just a ->
      let {
        (d, vpp2) = a;
      } in (Nat.plus_nat c d, x) : Polynomial.sumPolyVars vp1 vpp2;
  });
sumPolyVars [] vp2 = vp2;

polySum ::
  forall a.
    (Eq a) => Polynomial.Poly a -> Polynomial.Poly a -> Polynomial.Poly a;
polySum (Polynomial.Poly c vp1) (Polynomial.Poly d vp2) =
  Polynomial.Poly (Nat.plus_nat c d) (Polynomial.sumPolyVars vp1 vp2);

polyMult :: forall a. Nat.Nat -> Polynomial.Poly a -> Polynomial.Poly a;
polyMult n (Polynomial.Poly c vp) =
  Polynomial.Poly (Nat.times_nat n c)
    (List.mapa (\ cx -> (Nat.times_nat n (fst cx), snd cx)) vp);

checkVPLists ::
  forall a. (Eq a) => [(Nat.Nat, a)] -> [(Nat.Nat, a)] -> Util.Eithera a ();
checkVPLists vp ((n, x) : vq) =
  (if Nat.eq_nat n Nat.Zero_nat then Polynomial.checkVPLists vp vq
    else (case Polynomial.lookupVar x vp of {
           Nothing -> ErrorMonad.errora x;
           Just a ->
             let {
               (m, vpp) = a;
             } in ErrorMonad.bind (ErrorMonad.checkB (Nat.less_eq_nat n m) x)
                    (\ xa -> Polynomial.checkVPLists vpp vq);
         }));
checkVPLists uu [] = ErrorMonad.returna ();

checkPoly ::
  forall a.
    (Eq a,
      Show.Sshow a) => Polynomial.Poly a ->
                         Polynomial.Poly a -> Util.Eithera [[Char]] ();
checkPoly (Polynomial.Poly c vp) (Polynomial.Poly d vq) =
  ErrorMonad.bind
    (ErrorMonad.checkB (Nat.less_eq_nat d c)
      ["the constant of the left polynomial is too small"])
    (\ x ->
      ErrorMonad.catcha (Polynomial.checkVPLists vp vq)
        (\ xa ->
          ErrorMonad.errora
            [List.append (List.append "variable " (Show.sshow xa))
               " is problematic"]));

show_pair :: forall b. (Show.Sshow b) => (Nat.Nat, b) -> [Char];
show_pair (Nat.Suc (Nat.Suc n), v) =
  List.append (List.append (Show.sshow_nat (Nat.Suc (Nat.Suc n))) "*")
    (Show.sshow v);
show_pair (Nat.Suc Nat.Zero_nat, v) = Show.sshow v;
show_pair (Nat.Zero_nat, v) = [];

polyOfTerm ::
  forall a b.
    (Eq b) => (a -> (Nat.Nat, [Nat.Nat])) -> Term.Term a b -> Polynomial.Poly b;
polyOfTerm alpha (Term.Fun f ts) =
  let {
    a = alpha f;
    (c, list) = a;
  } in Polynomial.polySum (Polynomial.Poly c [])
         (List.foldra Polynomial.polySum
           (List.mapa
             (\ ct ->
               Polynomial.polyMult (fst ct)
                 (Polynomial.polyOfTerm alpha (snd ct)))
             (List.zipa list ts))
           (Polynomial.Poly Nat.Zero_nat []));
polyOfTerm alpha (Term.Var x) = Polynomial.Poly Nat.Zero_nat [(Nat.one_nat, x)];

merge_entries :: [[Char]] -> [Char];
merge_entries (x : xs) =
  List.append (List.append " + " x) (Polynomial.merge_entries xs);
merge_entries [] = [];

show_number :: Nat.Nat -> [Char];
show_number (Nat.Suc n) = Show.sshow_nat (Nat.Suc n);
show_number Nat.Zero_nat = [];

show_entries :: forall b. (Show.Sshow b) => Polynomial.Poly b -> [[Char]];
show_entries (Polynomial.Poly c vp) =
  List.filtera (\ entry -> not (entry == []))
    (Polynomial.show_number c : List.mapa Polynomial.show_pair vp);

sshow_poly :: forall b. (Show.Sshow b) => Polynomial.Poly b -> [Char];
sshow_poly p =
  (case Polynomial.show_entries p of {
    [] -> "0";
    d : es -> List.append d (Polynomial.merge_entries es);
  });

checkPolyGt ::
  forall v.
    (Eq v,
      Show.Sshow v) => Polynomial.Poly v ->
                         Polynomial.Poly v -> Util.Eithera [[Char]] ();
checkPolyGt p q =
  ErrorMonad.catcha
    (Polynomial.checkPoly p (Polynomial.polySum Polynomial.polyOne q))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Polynomial.sshow_poly p, " > ",
             Polynomial.sshow_poly q]
          : x));

checkSPolo ::
  forall f v.
    (Show.Sshow f, Eq v,
      Show.Sshow v) => (f -> (Nat.Nat, [Nat.Nat])) ->
                         (Term.Term f v, Term.Term f v) ->
                           Util.Eithera [[Char]] ();
checkSPolo alpha lr =
  ErrorMonad.catcha
    (Polynomial.checkPolyGt (Polynomial.polyOfTerm alpha (fst lr))
      (Polynomial.polyOfTerm alpha (snd lr)))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Term.sshow_term (fst lr), " > ",
             Term.sshow_term (snd lr)]
          : x));

checkPolyGe ::
  forall v.
    (Eq v,
      Show.Sshow v) => Polynomial.Poly v ->
                         Polynomial.Poly v -> Util.Eithera [[Char]] ();
checkPolyGe p q =
  ErrorMonad.catcha (Polynomial.checkPoly p q)
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Polynomial.sshow_poly p, " >= ",
             Polynomial.sshow_poly q]
          : x));

checkNSPolo ::
  forall f v.
    (Show.Sshow f, Eq v,
      Show.Sshow v) => (f -> (Nat.Nat, [Nat.Nat])) ->
                         (Term.Term f v, Term.Term f v) ->
                           Util.Eithera [[Char]] ();
checkNSPolo alpha lr =
  ErrorMonad.catcha
    (Polynomial.checkPolyGe (Polynomial.polyOfTerm alpha (fst lr))
      (Polynomial.polyOfTerm alpha (snd lr)))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Term.sshow_term (fst lr), " >= ",
             Term.sshow_term (snd lr)]
          : x));

showInterpretation :: forall a. (a -> (Nat.Nat, [Nat.Nat])) -> a -> [Char];
showInterpretation alpha f =
  Polynomial.sshow_poly
    (Polynomial.polyOfTerm alpha
      (Polynomial.genTerm f (List.size_list (snd (alpha f)))));

createRepresentation ::
  forall a.
    (Eq a, Show.Sshow a) => (a -> (Nat.Nat, [Nat.Nat])) -> [a] -> [[Char]];
createRepresentation alpha fs =
  "polynomial interpretation: " :
    List.mapa
      (\ g ->
        List.concata
          ["Pol(", Show.sshow g, ") = ", Polynomial.showInterpretation alpha g])
      (List.remdups fs);

createPolyInter ::
  forall a. (Eq a) => [(a, (Nat.Nat, [Nat.Nat]))] -> a -> (Nat.Nat, [Nat.Nat]);
createPolyInter ((g, i) : rest) f =
  (if HOL.eqop f g then i else Polynomial.createPolyInter rest f);
createPolyInter [] f = (Nat.Zero_nat, []);

createRedPairProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Util.Eithera [[Char]] ();
createRedPairProc (Polynomial.Polo interpr) p rP r =
  let {
    alpha = Polynomial.createPolyInter interpr;
  } in RedPair.checkRedPairProc (Polynomial.checkSPolo alpha)
         (Polynomial.checkNSPolo alpha) "polynomial order"
         (Polynomial.createRepresentation alpha
           (List.remdups (List.mapa fst interpr)))
         p rP r;

}
